/*****************************************************************************/
/*  file_transfer.c - contains the ui and network routines for file transfer */
/*  Copyright (C) 1998-2000 Brian Masney <masneyb@seul.org>                  */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either version 2 of the License, or        */
/*  (at your option) any later version.                                      */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111 USA      */
/*****************************************************************************/

#include "ftp.h"

struct recursive_transfer
{
  struct ftp_transfer_data *tdata;
  GList *current_file;
  char *path;
  char *dest;
  char *root_path;
  char *start_file;
};

static void get_local_files_and_dirs 	( struct recursive_transfer *transdata );
static void asktrans 			( struct ftp_transfer_data *tdata );
static void trans_selectall 		( GtkWidget * widget,
					  struct ftp_transfer_data *tdata );
static void trans_unselectall 		( GtkWidget * widget,
					  struct ftp_transfer_data *tdata );
static void overwrite 			( GtkWidget * widget, 
					  struct ftp_transfer_data *tdata );
static void resume 			( GtkWidget * widget, 
					  struct ftp_transfer_data *tdata );
static void skip 			( GtkWidget * widget, 
					  struct ftp_transfer_data *tdata );
static void ok 				( GtkWidget * widget, 
					  struct ftp_transfer_data *tdata );
static void cancel 			( GtkWidget * widget, 
					  struct ftp_transfer_data *tdata );
static void calc_kbs 			( struct ftp_transfer_data *tdata, 
					  ssize_t num_read );
static void set_file_transfer_mode 	( struct ftp_transfer_data *tdata );
static void goto_next_file 		( struct ftp_transfer_data *tdata );
static int get_status 			( struct ftp_transfer_data *tdata, 
					  ssize_t num_read );
static mode_t parse_attribs 		( char *attribs );
static char *convert_ascii 		( char *buf, 
					  ssize_t *len, 
					  int direction );

static GtkWidget *dialog = NULL, *status;
static int skip_same_state;

void
retrCB (GtkWidget * widget, gpointer data)
{
  struct ftp_transfer_data *tdata;

  if (!check_status (_("Retrieve Files"), &window2, 0, 1,
       window1.hdata->ftpdata->put_file != NULL
       && window2.hdata->ftpdata->get_file != NULL))
    return;

  if (!IS_CONNECTED (&window1))
    {
      ftp_log (gftp_logging_misc, NULL,
	       _("Retrieve Files: Not connected to a remote site\n"));
      return;
    }

  if ((tdata = inittrans (1, 1, window2.hdata->files)) != NULL)
    dotrans (tdata);
}


void
putCB (GtkWidget * widget, gpointer data)
{
  struct ftp_transfer_data *tdata;

  if (!check_status (_("Put Files"), &window1, 0, 1,
       window1.hdata->ftpdata->put_file != NULL
       && window2.hdata->ftpdata->get_file != NULL))
    return;

  if (!IS_CONNECTED (&window2))
    {
      ftp_log (gftp_logging_misc, NULL,
	       _("Put Files: Not connected to a remote site\n"));
      return;
    }

  if ((tdata = inittrans (0, 1, window1.hdata->files)) != NULL)
    dotrans (tdata);
}


struct ftp_transfer_data *
inittrans (int direction, int add_to_queue, GList * files)
{
  struct ftp_window_data *fromwdata, *towdata;
  struct ftp_file_data *newfle, *tempfle;
  struct ftp_transfer_data *tdata;
  GList *templist, *filelist;
  int curpos;

  skip_same_state = 0;
  tdata = new_tdata ();
  if (direction)
    {
      tdata->direction = 1;
      fromwdata = &window2;
      towdata = &window1;
    }
  else
    {
      fromwdata = &window1;
      towdata = &window2;
    }
  copy_hdata_struct (fromwdata->hdata, tdata->fromhdata);
  copy_hdata_struct (towdata->hdata, tdata->tohdata);
  tdata->fromhdata->wdata = fromwdata;
  tdata->tohdata->wdata = towdata;
  fromwdata->hdata->ftpdata->init (tdata->fromhdata->ftpdata);
  towdata->hdata->ftpdata->init (tdata->tohdata->ftpdata);

  tdata->fromhdata->totalfiles = 0;
  curpos = 0;
  filelist = files;
  templist = GTK_CLIST (tdata->fromhdata->wdata->listbox)->selection;
  while (templist != NULL)
    {
      templist = get_next_selection (templist, &filelist, &curpos);
      tempfle = (struct ftp_file_data *) filelist->data;
      if (strcmp (tempfle->file, "..") == 0)
	{
	  ftp_log (gftp_logging_misc, NULL,
		   _("Transfer Files: Skipping the transfer of the .. directory\n"));
	  continue;
	}
      else if (tempfle->isdir)
        add_entire_directory (tdata, tempfle, 
			      GFTP_GET_DIRECTORY (fromwdata->hdata->ftpdata), 
			      GFTP_GET_DIRECTORY (towdata->hdata->ftpdata),
			      fromwdata->hdata->ftpdata);
      else
	{
	  newfle = g_malloc0 (sizeof (struct ftp_file_data));
	  copy_fdata_struct (tempfle, newfle);

	  if (newfle->file)
	    g_free (newfle->file);
	  if (newfle->remote_file)
	    g_free (newfle->remote_file);
	  newfle->file =
	    g_strconcat (GFTP_GET_DIRECTORY (fromwdata->hdata->ftpdata), "/",
			 tempfle->file, NULL);
	  remove_double_slashes (newfle->file);
	  newfle->remote_file = g_strconcat (
                                  GFTP_GET_DIRECTORY (towdata->hdata->ftpdata),
                                  "/", tempfle->file, NULL);
	  remove_double_slashes (newfle->remote_file);

	  newfle->ascii = get_file_transfer_mode (newfle->file);
	  tdata->fromhdata->totalfiles++;
	  tdata->fromhdata->files = g_list_append (tdata->fromhdata->files, 
                                                   newfle);
	}
    }
  done_with_dirs (fromwdata->hdata->ftpdata);
  gftp_set_directory (fromwdata->hdata->ftpdata, 
		      GFTP_GET_DIRECTORY (fromwdata->hdata->ftpdata));

  if (tdata->fromhdata->files == NULL)
    {
      free_tdata (tdata);
      return (NULL);
    }
  tdata->curfle = tdata->fromhdata->files;
  if (add_to_queue)
    add_file_transfer (tdata);
  return (tdata);
}


void
add_entire_directory (struct ftp_transfer_data *tdata, 
		      struct ftp_file_data *tempfle,
		      char *fromdir, char *todir,
		      gftp_request * fromreq)
{
  FILE *oldsockfd, *oldsockfd_write, *olddatafd;
  struct recursive_transfer transdata;
  char *tempstr, *temp1str, *pos;
  GtkWidget *tempwid, *vbox;

  if (dialog == NULL)
    {
      dialog = gtk_window_new (GTK_WINDOW_DIALOG);
      gtk_grab_add (dialog);
      gtk_window_set_title (GTK_WINDOW (dialog),
			    _("Getting directory listings"));
      gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
      gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
				 GTK_SIGNAL_FUNC (gtk_widget_destroy),
				 GTK_OBJECT (dialog));

      vbox = gtk_vbox_new (FALSE, 5);
      gtk_container_border_width (GTK_CONTAINER (vbox), 10);
      gtk_container_add (GTK_CONTAINER (dialog), vbox);
      gtk_widget_show (vbox);

      tempwid = gtk_label_new ( _("Please wait while getting directory listings"));
      gtk_box_pack_start (GTK_BOX (vbox), tempwid, TRUE, TRUE, 0);
      gtk_widget_show (tempwid);

      status = gtk_label_new ("");
      gtk_box_pack_start (GTK_BOX (vbox), status, TRUE, TRUE, 0);
      gtk_widget_show (status);
      gtk_widget_show (dialog);
      fix_display ();
    }

  if ((pos = strrchr (tempfle->file, '/')) == NULL)
    pos = tempfle->file;
  else
    pos++;
  memset (&transdata, 0, sizeof (struct recursive_transfer));
  transdata.tdata = tdata;
  if (*fromdir == '\0')
    {
      tempstr = g_malloc (strlen (tempfle->file) + 1);
      strcpy (tempstr, tempfle->file);
    }
  else
    tempstr = g_strconcat (fromdir, "/", pos, NULL);
  temp1str = g_strconcat (todir, "/", pos, NULL);
  transdata.path = tempstr;
  transdata.dest = temp1str;
  transdata.root_path = g_malloc (strlen (transdata.path) + 1);
  strcpy (transdata.root_path, transdata.path);
  transdata.start_file = NULL;

  oldsockfd = tdata->fromhdata->ftpdata->sockfd;
  oldsockfd_write = tdata->fromhdata->ftpdata->sockfd_write;
  olddatafd = tdata->fromhdata->ftpdata->datafd;

  tdata->fromhdata->ftpdata->sockfd = fromreq->sockfd;
  tdata->fromhdata->ftpdata->sockfd_write = fromreq->sockfd_write;
  tdata->fromhdata->ftpdata->datafd = fromreq->datafd;

  get_local_files_and_dirs (&transdata);

  tdata->fromhdata->ftpdata->sockfd = oldsockfd;
  tdata->fromhdata->ftpdata->sockfd_write = oldsockfd_write;
  tdata->fromhdata->ftpdata->datafd = olddatafd;

  g_free (tempstr);
  g_free (temp1str);
}


void
done_with_dirs (void)
{
  if (dialog != NULL)
    {
      gtk_widget_destroy (dialog);
      dialog = NULL;
    }
}


static void
get_local_files_and_dirs (struct recursive_transfer *transdata)
{
  struct ftp_file_data *tempfle;
  GList *templist;
  char *tempstr;
  int i;

  for (i = 0; i < transdata->tdata->num_dirs_to_be_made; i++)
    {
      if (strcmp (transdata->tdata->dirs_to_be_made[i], transdata->dest) == 0)
	break;
    }

  if (i == transdata->tdata->num_dirs_to_be_made)
    {
      transdata->tdata->num_dirs_to_be_made++;
      transdata->tdata->dirs_to_be_made = g_realloc (
                   transdata->tdata->dirs_to_be_made,
		   transdata->tdata->num_dirs_to_be_made * sizeof (char *));
      transdata->tdata->dirs_to_be_made
                          [transdata->tdata->num_dirs_to_be_made - 1] =
                          g_malloc (strlen (transdata->dest) + 1);
      strcpy (transdata->tdata->
	      dirs_to_be_made[transdata->tdata->num_dirs_to_be_made - 1],
	      transdata->dest);
    }

  templist = get_remote_files (transdata->tdata->fromhdata, transdata->path, 
                               &i, 1, GTK_LABEL (status));

  if (templist != NULL)
    {
      tempfle = (struct ftp_file_data *) templist->data;
      free_fdata (tempfle);
      templist = g_list_remove_link (templist, templist);
    }

  if (transdata->tdata->fromhdata->files == NULL)
    transdata->tdata->fromhdata->files = templist;
  else
    transdata->tdata->fromhdata->files = g_list_concat (
                                transdata->tdata->fromhdata->files, templist);

  if (transdata->current_file == NULL)
    transdata->current_file = templist;

  while (templist != NULL)
    {
      tempfle = (struct ftp_file_data *) templist->data;
      tempstr = tempfle->file;
      tempfle->file = g_strconcat (transdata->path, "/", tempstr, NULL);
      tempfle->remote_file = g_strconcat (transdata->dest, "/", tempstr, NULL);
      g_free (tempstr);
      tempfle->remote_size = tempfle->size;
      templist = templist->next;
    }

  while (transdata->current_file != NULL)
    {
      tempfle = (struct ftp_file_data *) transdata->current_file->data;
      if (tempfle->isdir)
	{
	  transdata->path = tempfle->file;
	  transdata->dest = tempfle->remote_file;
	  templist = transdata->current_file;
	  transdata->current_file = transdata->current_file->next;
	  transdata->tdata->fromhdata->files = g_list_remove_link (
                                transdata->tdata->fromhdata->files, templist);
	  get_local_files_and_dirs (transdata);
	  return;
	}
      else
	{
	  transdata->tdata->fromhdata->totalfiles++;
	  tempfle->ascii = get_file_transfer_mode (tempfle->file);
	}
      transdata->current_file = transdata->current_file->next;
    }
}


void
dotrans (struct ftp_transfer_data *tdata)
{
  struct ftp_window_data *wdatafrom, *wdatato;
  struct ftp_file_data *tempfle;
  int same, swapped;
  long size;

  if (tdata->direction)
    {
      wdatafrom = &window2;
      wdatato = &window1;
    }
  else
    {
      wdatafrom = &window1;
      wdatato = &window2;
    }

  if (compare_hdata_structs (tdata->tohdata->ftpdata, 
                             wdatato->hdata->ftpdata, 0))
    {
      swap_socks (tdata->fromhdata->ftpdata, wdatafrom->hdata->ftpdata);
      swap_socks (tdata->tohdata->ftpdata, wdatato->hdata->ftpdata);
      swapped = 1;
    }
  else
    swapped = 0;

  same = 0;
  pthread_mutex_lock (&tdata->mutex);
  while (tdata->curfle != NULL)
    {
      tempfle = (struct ftp_file_data *) tdata->curfle->data;
      /* FIXME - do some better checking here */
      if ((size = gftp_get_file_size (tdata->tohdata->ftpdata,
			              tempfle->remote_file)) > 0)
	{
	  tempfle->remote_size = size;
	  tempfle->exists = 1;
	  same = 1;
	}
      tdata->curfle = tdata->curfle->next;
    }

  if (swapped)
    {
      swap_socks (wdatafrom->hdata->ftpdata, tdata->fromhdata->ftpdata);
      swap_socks (wdatato->hdata->ftpdata, tdata->tohdata->ftpdata);
    }

  if (same == 0)
    {
      tdata->show = 1;
      tdata->ready = 1;
      pthread_mutex_unlock (&tdata->mutex);
    }
  else
    {
      pthread_mutex_unlock (&tdata->mutex);
      asktrans (tdata);
    }
}


static void
asktrans (struct ftp_transfer_data *tdata)
{
  GtkWidget *tempwid, *scroll, *hbox;
  char *dltitles[4], *add_data[4] = { NULL, NULL, NULL, NULL };
  struct ftp_file_data *tempfle;
  char *tempstr, *pos;
  GList *templist;
  size_t len;
  int i;

  dltitles[0] = _("Filename");
  dltitles[1] = _("Local Size");
  dltitles[2] = _("Remote Size");
  dltitles[3] = _("Action");
  dialog = gtk_dialog_new ();
  gtk_grab_add (dialog);
  gtk_window_set_title (GTK_WINDOW (dialog), tdata->direction 
                        ? _("Download Files") : _("Upload Files"));
  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
  gtk_container_border_width (GTK_CONTAINER
			      (GTK_DIALOG (dialog)->action_area), 5);
  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 5);
  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->action_area), 35);
  gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
  gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
                             GTK_SIGNAL_FUNC (gtk_widget_destroy),
                             GTK_OBJECT (dialog));

  tempwid = gtk_label_new (
		   _("The following file(s) exist on both the local and remote computer\nPlease select what you would like to do"));
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), tempwid, FALSE,
		      FALSE, 0);
  gtk_widget_show (tempwid);

  scroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_set_usize (scroll, 450, 200);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  tdata->clist = gtk_clist_new_with_titles (4, dltitles);
  gtk_container_add (GTK_CONTAINER (scroll), tdata->clist);
  gtk_clist_set_selection_mode (GTK_CLIST (tdata->clist),
				GTK_SELECTION_EXTENDED);
  gtk_clist_set_column_width (GTK_CLIST (tdata->clist), 0, 100);
  gtk_clist_set_column_justification (GTK_CLIST (tdata->clist), 1,
				      GTK_JUSTIFY_RIGHT);
  gtk_clist_set_column_width (GTK_CLIST (tdata->clist), 1, 85);
  gtk_clist_set_column_justification (GTK_CLIST (tdata->clist), 2,
				      GTK_JUSTIFY_RIGHT);
  gtk_clist_set_column_width (GTK_CLIST (tdata->clist), 2, 85);
  gtk_clist_set_column_width (GTK_CLIST (tdata->clist), 3, 85);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), scroll, TRUE, TRUE,
		      0);
  gtk_widget_show (tdata->clist);
  gtk_widget_show (scroll);

  templist = tdata->fromhdata->files;
  while (templist != NULL)
    {
      tempfle = (struct ftp_file_data *) templist->data;
      if (tempfle->exists)
	{
	  pos = tempfle->remote_file;
	  len = strlen (GFTP_GET_DIRECTORY (tdata->tohdata->ftpdata));
	  if (strncmp (pos, GFTP_GET_DIRECTORY (tdata->tohdata->ftpdata), len)
	      == 0)
	    pos = tempfle->remote_file + len + 1;
	  add_data[0] = pos;
	  add_data[3] = _("Overwrite");
	  i = gtk_clist_append (GTK_CLIST (tdata->clist), add_data);

	  tempstr = insert_commas (tempfle->size);
	  gtk_clist_set_text (GTK_CLIST (tdata->clist), i, 
                              tdata->direction ? 2 : 1, tempstr);
	  g_free (tempstr);

	  tempstr = insert_commas (tempfle->remote_size);
	  gtk_clist_set_text (GTK_CLIST (tdata->clist), i, 
                              tdata->direction ? 1 : 2, tempstr);
	  gtk_clist_set_row_data (GTK_CLIST (tdata->clist), i, (void *) i);
	  g_free (tempstr);
	}
      templist = templist->next;
    }

  gtk_clist_select_all (GTK_CLIST (tdata->clist));

  hbox = gtk_hbox_new (TRUE, 20);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);

  tempwid = gtk_button_new_with_label (_("Overwrite"));
  gtk_box_pack_start (GTK_BOX (hbox), tempwid, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
		      GTK_SIGNAL_FUNC (overwrite), (gpointer) tdata);
  gtk_widget_show (tempwid);

  tempwid = gtk_button_new_with_label (_("Resume"));
  gtk_box_pack_start (GTK_BOX (hbox), tempwid, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
		      GTK_SIGNAL_FUNC (resume), (gpointer) tdata);
  gtk_widget_show (tempwid);

  tempwid = gtk_button_new_with_label (_("Skip File"));
  gtk_box_pack_start (GTK_BOX (hbox), tempwid, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (skip),
		      (gpointer) tdata);
  gtk_widget_show (tempwid);

  hbox = gtk_hbox_new (TRUE, 20);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), hbox, TRUE, TRUE, 0);
  gtk_widget_show (hbox);

  tempwid = gtk_button_new_with_label (_("Select All"));
  gtk_box_pack_start (GTK_BOX (hbox), tempwid, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
		      GTK_SIGNAL_FUNC (trans_selectall), (gpointer) tdata);
  gtk_widget_show (tempwid);

  tempwid = gtk_button_new_with_label (_("Deselect All"));
  gtk_box_pack_start (GTK_BOX (hbox), tempwid, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
		      GTK_SIGNAL_FUNC (trans_unselectall), (gpointer) tdata);
  gtk_widget_show (tempwid);

  tempwid = gtk_button_new_with_label (_("OK"));
  GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
		      TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked", GTK_SIGNAL_FUNC (ok),
		      (gpointer) tdata);
  gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_widget_grab_default (tempwid);
  gtk_widget_show (tempwid);

  tempwid = gtk_button_new_with_label (_("  Cancel  "));
  GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
		      TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (tempwid), "clicked",
		      GTK_SIGNAL_FUNC (cancel), (gpointer) tdata);
  gtk_signal_connect_object (GTK_OBJECT (tempwid), "clicked",
			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
			     GTK_OBJECT (dialog));
  gtk_widget_show (tempwid);

  gtk_widget_show (dialog);
  dialog = NULL;
}


static void
trans_selectall (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  gtk_clist_select_all (GTK_CLIST (tdata->clist));
}


static void
trans_unselectall (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  gtk_clist_unselect_all (GTK_CLIST (tdata->clist));
}


static void
overwrite (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  struct ftp_file_data *tempfle;
  GList *templist, *filelist;
  int curpos;

  curpos = 0;
  filelist = tdata->fromhdata->files;
  templist = GTK_CLIST (tdata->clist)->selection;
  while (templist != NULL)
    {
      templist = get_next_selection (templist, &filelist, &curpos);
      tempfle = (struct ftp_file_data *) filelist->data;
      if (tempfle->exists)
	{
	  tempfle->restart = 0;
	  tempfle->skip = 0;
	  gtk_clist_set_text (GTK_CLIST (tdata->clist), curpos, 3,
			      _("Overwrite"));
	}
    }
}


static void
resume (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  struct ftp_file_data *tempfle;
  GList *templist, *filelist;
  int curpos;

  curpos = 0;
  filelist = tdata->fromhdata->files;
  templist = GTK_CLIST (tdata->clist)->selection;
  while (templist != NULL)
    {
      templist = get_next_selection (templist, &filelist, &curpos);
      tempfle = (struct ftp_file_data *) filelist->data;
      if (tempfle->exists)
	{
	  tempfle->restart = 1;
	  tempfle->skip = 0;
	  gtk_clist_set_text (GTK_CLIST (tdata->clist), curpos, 3,
			      _("Resume"));
	}
    }
}


static void
skip (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  struct ftp_file_data *tempfle;
  GList *templist, *filelist;
  int curpos;

  curpos = 0;
  filelist = tdata->fromhdata->files;
  templist = GTK_CLIST (tdata->clist)->selection;
  while (templist != NULL)
    {
      templist = get_next_selection (templist, &filelist, &curpos);
      tempfle = (struct ftp_file_data *) filelist->data;
      if (tempfle->exists)
	{
	  tempfle->restart = 0;
	  tempfle->skip = 1;
	  gtk_clist_set_text (GTK_CLIST (tdata->clist), curpos, 3, _("Skip"));
	}
    }
}


static void
ok (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  struct ftp_file_data *tempfle;
  GList *templist, *dellist;

  pthread_mutex_lock (&tdata->mutex);
  templist = tdata->fromhdata->files;
  while (templist != NULL)
    {
      tempfle = (struct ftp_file_data *) templist->data;
      if (tempfle->skip)
	{
	  free_fdata (tempfle);
	  dellist = templist;
	  templist = templist->next;
	  tdata->fromhdata->files =
	    g_list_remove_link (tdata->fromhdata->files, dellist);
	  tdata->fromhdata->totalfiles--;
	  continue;
	}
      templist = templist->next;
    }
  if (tdata->fromhdata->files == NULL)
    tdata->done = 1;
  else
    tdata->show = tdata->ready = 1;
  pthread_mutex_unlock (&tdata->mutex);
}


static void
cancel (GtkWidget * widget, struct ftp_transfer_data *tdata)
{
  pthread_mutex_lock (&tdata->mutex);
  tdata->show = 0;
  tdata->done = tdata->ready = 1;
  pthread_mutex_unlock (&tdata->mutex);
}


void *
transfer_files (void *ptr)
{
  struct ftp_transfer_data *tdata;
  struct ftp_file_data *tempfle;
  char buf[8192], *tempstr;
  long fromsize, tosize;
  ssize_t num_read;
  int i, success;

  tdata = (struct ftp_transfer_data *) ptr;
  pthread_detach (pthread_self ());
  pthread_mutex_lock (&tdata->mutex);
  gettimeofday (&tdata->inittime, NULL);
  pthread_mutex_unlock (&tdata->mutex);
  gftp_logging (tdata->fromhdata->ftpdata, 1, queue_log, NULL);
  gftp_logging (tdata->tohdata->ftpdata, 1, queue_log, NULL);
  if (ftp_connect (tdata->fromhdata, 0) && ftp_connect (tdata->tohdata, 0))
    {
      for (i = 0; i < tdata->num_dirs_to_be_made; i++)
	{
	  gftp_make_directory (tdata->tohdata->ftpdata,
			       tdata->dirs_to_be_made[i]);
	  g_free (tdata->dirs_to_be_made[i]);
	}
      g_free (tdata->dirs_to_be_made);

      while (!tdata->cancel && tdata->curfle != NULL)
	{
	  tempfle = (struct ftp_file_data *) tdata->curfle->data;
	  success = 1;
          fromsize = tosize = 0;
	  set_file_transfer_mode (tdata);

	  if (tempfle->restart)
            tosize = gftp_get_file_size (tdata->tohdata->ftpdata, 
                                         tempfle->remote_file);

	  fromsize = gftp_transfer_file (tdata->fromhdata->ftpdata, 
                                tempfle->file, 0, tdata->tohdata->ftpdata,
				tempfle->remote_file, tosize);
	  if (fromsize < 0)
	    {
	      goto_next_file (tdata);
	      continue;
	    }

	  pthread_mutex_lock (&tdata->mutex);
          tdata->total_bytes = tdata->total_bytes - tempfle->remote_size 
                               + fromsize;
	  tdata->current_file_retries++;
	  tdata->current_file_number++;
	  if (fromsize != 0)
	    tempfle->remote_size = fromsize;
	  gettimeofday (&tdata->starttime, NULL);
	  memcpy (&tdata->lasttime, &tdata->starttime,
		  sizeof (tdata->lasttime));
	  tdata->curtrans = tosize;
	  tdata->curmax = tempfle->remote_size == 0 ? tempfle->size : 
                                                      tempfle->remote_size;
	  tdata->trans_bytes += tosize;
	  tdata->sttrans = tosize;
	  pthread_mutex_unlock (&tdata->mutex);

	  num_read = 0;
          if (GFTP_GET_DATA_FD (tdata->fromhdata->ftpdata) != NULL)
            {
	      while (!tdata->cancel && 
                    (num_read = gftp_get_next_file_chunk (
                             tdata->fromhdata->ftpdata, buf, sizeof (buf))) > 0)
	        {
	          calc_kbs (tdata, num_read);

                  if (GFTP_GET_DATA_TYPE (tdata->fromhdata->ftpdata) 
                      == GFTP_TYPE_ASCII)
                    tempstr = convert_ascii (buf, &num_read, tdata->direction);
                  else
                    tempstr = buf;

                  if (gftp_put_next_file_chunk (tdata->tohdata->ftpdata, 
                      tempstr, num_read) < 0)
                    break; 

                  if (GFTP_GET_DATA_TYPE (tdata->fromhdata->ftpdata) 
                      == GFTP_TYPE_ASCII && !tdata->direction)
                     g_free (tempstr);
	        }
              gftp_put_next_file_chunk (tdata->tohdata->ftpdata, NULL, 0);
            }
          else
            tdata->fxp = 1;

          if (num_read < 0)
            {
              gftp_disconnect (tdata->fromhdata->ftpdata); 
              gftp_disconnect (tdata->tohdata->ftpdata);
            }
          else
            {
	      gftp_end_transfer (tdata->fromhdata->ftpdata);
	      gftp_end_transfer (tdata->tohdata->ftpdata);
            }

	  if (success && !tdata->cancel)
	    {
	      switch (get_status (tdata, num_read))
		{
		case -1:
		  success = 0;
		case 1:
		  continue;
		default:
		  success = 1;
		}
	    }

	  if (success && !tdata->cancel)
	    {
	      queue_log (gftp_logging_misc, NULL,
			 _("Successfully transferred %s\n"),
			 tempfle->file);
	    }
	  else
	    {
	      queue_log (gftp_logging_error, NULL,
			 _("Could not download %s from %s\n"),
			 tempfle->remote_file,
			 GFTP_GET_HOSTNAME (tdata->fromhdata->ftpdata));
	      gftp_disconnect (tdata->tohdata->ftpdata);
	      gftp_disconnect (tdata->fromhdata->ftpdata);
	      break;
	    }

	  if (preserve_permissions)
	    {
	      if (tempfle->attribs)
		gftp_chmod (tdata->tohdata->ftpdata, tempfle->remote_file,
			    parse_attribs (tempfle->attribs));
	      gftp_set_file_time (tdata->tohdata->ftpdata,
				  tempfle->remote_file, tempfle->datetime);
	    }
	  goto_next_file (tdata);
	}
    }

  pthread_mutex_lock (&tdata->mutex);
  tdata->done = 1;
  pthread_mutex_unlock (&tdata->mutex);
  pthread_exit (NULL);
}


static void
calc_kbs (struct ftp_transfer_data *tdata, ssize_t num_read)
{
  unsigned long waitusecs, difftime;
  struct timeval tv;
  float curkbs;

  gettimeofday (&tv, NULL);
  pthread_mutex_lock (&tdata->mutex);
  difftime = (tv.tv_sec - tdata->starttime.tv_sec) + ((tv.tv_usec -
             tdata->starttime.tv_usec) / 1000000);
  if (difftime == 0)
    tdata->kbs = 0;
  else
    tdata->kbs = ((float) (tdata->curtrans - tdata->sttrans) / 1024) /
                 (float) difftime;
  tdata->curtrans += num_read;
  tdata->trans_bytes += num_read;
  tdata->stalled = 0;
  pthread_mutex_unlock (&tdata->mutex);

  difftime = (tv.tv_sec - tdata->lasttime.tv_sec) + ((tv.tv_usec -
             tdata->lasttime.tv_usec) / 1000000);
  if (difftime == 0)
    curkbs = (float) num_read / 1024;
  else
    curkbs = ((float) num_read / 1024) / (float) difftime;

  if (tdata->maxkbs > 0 && curkbs > tdata->maxkbs)
    {
      waitusecs = ((float) num_read / 1024) / tdata->maxkbs * 1000000 - 
                  difftime;
      usleep (waitusecs);
    }
  gettimeofday (&tdata->lasttime, NULL);
}


static void
set_file_transfer_mode (struct ftp_transfer_data *tdata)
{
  if (((struct ftp_file_data *) tdata->curfle->data)->ascii)
    gftp_set_data_type (tdata->fromhdata->ftpdata, GFTP_TYPE_ASCII);
  else
    gftp_set_data_type (tdata->fromhdata->ftpdata, GFTP_TYPE_BINARY);
}


static void
goto_next_file (struct ftp_transfer_data *tdata)
{
  pthread_mutex_lock (&tdata->mutex);
  tdata->next_file = 1;
  tdata->curfle = tdata->curfle->next;
  tdata->current_file_retries = 0;
  pthread_mutex_unlock (&tdata->mutex);
}


static int
get_status (struct ftp_transfer_data *tdata, ssize_t num_read)
{
  struct ftp_file_data *tempfle;
  struct timeval tv;

  tempfle = (struct ftp_file_data *) tdata->curfle->data;

  if (num_read < 0 || tdata->curtrans < tdata->curmax)
    {
      if (tdata->current_file_retries >= retries)
	{
	  queue_log (gftp_logging_error, NULL,
		     _("Error: Remote site %s disconnected. Max retries reached...giving up\n"),
		     GFTP_GET_HOSTNAME (tdata->fromhdata->ftpdata));
	  return (-1);
	}
      else
	{
	  queue_log (gftp_logging_error, NULL,
		     _("Error: Remote site %s disconnected. Will reconnect in %d seconds\n"),
		     GFTP_GET_HOSTNAME (tdata->fromhdata->ftpdata),
		     sleep_time);
	}
      gftp_disconnect (tdata->fromhdata->ftpdata);
      tv.tv_sec = sleep_time;
      tv.tv_usec = 0;
      select (0, NULL, NULL, NULL, &tv);
      if (ftp_connect (tdata->fromhdata, 0)
	  && ftp_connect (tdata->tohdata, 0))
	{
          if (GFTP_GET_DATA_FD (tdata->tohdata->ftpdata))
   	    fclose (GFTP_GET_DATA_FD (tdata->tohdata->ftpdata));
          if (GFTP_GET_DATA_FD (tdata->fromhdata->ftpdata))
	    fclose (GFTP_GET_DATA_FD (tdata->fromhdata->ftpdata));
	  pthread_mutex_lock (&tdata->mutex);
	  tempfle->restart = 1;
	  tempfle->size = tdata->curtrans;
	  pthread_mutex_unlock (&tdata->mutex);
	  tdata->current_file_number--;	/* Decrement this because it will 
					   be incremented when we continue in the loop */
	  return (1);
	}
    }

  return (0);
}


static mode_t
parse_attribs (char *attribs)
{
  mode_t mode;
  int cur;

  mode = 0;
  cur = 0;
  if (tolower (attribs[3]) == 's')
    cur += 4;
  if (tolower (attribs[6]) == 's')
    cur += 2;
  mode = (mode * 10) + cur;

  cur = 0;
  if (attribs[1] == 'r')
    cur += 4;
  if (attribs[2] == 'w')
    cur += 2;
  if (attribs[3] == 'x')
    cur += 1;
  mode = (mode * 10) + cur;

  cur = 0;
  if (attribs[4] == 'r')
    cur += 4;
  if (attribs[5] == 'w')
    cur += 2;
  if (attribs[6] == 'x')
    cur += 1;
  mode = (mode * 10) + cur;

  cur = 0;
  if (attribs[7] == 'r')
    cur += 4;
  if (attribs[8] == 'w')
    cur += 2;
  if (attribs[9] == 'x')
    cur += 1;
  mode = (mode * 10) + cur;

  return (mode);
}

static char * 
convert_ascii (char *buf, ssize_t *len, int direction)
{
  ssize_t i, j, newsize;
  char *tempstr;

  if (direction)
    {
      for (i = 0, j = 0; i < *len; i++)
        {
          if (buf[i] != '\r')
            buf[j++] = buf[i];
        }
      tempstr = buf;
    }
  else
    {
      newsize = 0;
      for (i = 0; i < *len; i++)
        {
          newsize++;
          if (i > 0 && buf[i] == '\n' && buf[i - 1] != '\r')
            newsize++;
        }

      tempstr = g_malloc (newsize);

      for (i = 0, j = 0; i < *len; i++)
        {
          if (i > 0 && buf[i] == '\n' && buf[i - 1] != '\r')
            tempstr[j++] = '\r';
          tempstr[j++] = buf[i];
        }
      *len = newsize;
    }
  return (tempstr);
}

